#![allow(dead_code)]

use std::{collections::BTreeMap, fmt::Debug, sync::Arc};

use crate::{
    buffers::{Buffer, BufferDesc},
    commands, def_monotonic_id,
    device::{self, DeviceOwned},
    images::{self, Image, ImageDesc, SUBRESOURCERANGE_COLOR_ALL},
    texture,
    util::{self, Rgba},
    EguiState, SwapchainFrame,
};
use ash::{prelude::VkResult, vk};
use egui::Color32;
use itertools::Itertools;
use petgraph::visit::NodeRef;

def_monotonic_id!(pub RenderGraphResourceId);

#[derive(Debug, Clone)]
pub enum RenderGraphResourceDesc {
    Image(ImageDesc),
    Buffer(BufferDesc),
}

#[derive(Debug)]
pub enum RenderGraphResource {
    Framebuffer(Arc<SwapchainFrame>),
    ImportedImage(Arc<Image>),
    ImportedBuffer(Arc<Buffer>),
    Image(Image),
    Buffer(Buffer),
}

#[derive(Debug, Clone, Copy)]
pub enum LoadOp {
    Clear(Rgba),
    Load,
    DontCare,
}

#[derive(Debug, Clone, Copy)]
pub enum StoreOp {
    DontCare,
    Store,
}

struct AttachmentInfo {
    size: glam::UVec2,
    format: vk::Format,
    load: LoadOp,
    store: StoreOp,
}

pub struct RenderContext<'a> {
    device: device::Device,
    cmd: commands::SingleUseCommand,
    resources: &'a BTreeMap<RenderGraphResourceId, RenderGraphResource>,
}

#[derive(Debug, Clone, Copy, Default)]
pub struct ResourceAccess {
    stage: vk::PipelineStageFlags2,
    mask: vk::AccessFlags2,
    layout: Option<vk::ImageLayout>,
}

impl ResourceAccess {
    fn undefined() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::UNDEFINED),
        }
    }
}

pub struct EguiPrePass {
    out_resources:
        BTreeMap<egui::TextureId, (RenderGraphResourceId, texture::TextureId, ResourceAccess)>,
    staging_image: Arc<Image>,
    staging_buffer: Arc<Buffer>,
    aliased_images: BTreeMap<egui::TextureId, (usize, usize, [usize; 2], Arc<Image>)>,
    tessellated: Vec<egui::ClippedPrimitive>,
    texture_data: BTreeMap<egui::TextureId, egui::epaint::ImageDelta>,
}

impl Debug for EguiPrePass {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("EguiPrePass")
            .field("out_resources", &self.out_resources)
            .finish_non_exhaustive()
    }
}

impl EguiPrePass {
    pub fn new(
        dev: &device::Device,
        rg: &mut RenderGraph,
        textures: &mut crate::texture::TextureManager,
        egui_state: &mut EguiState,
        egui: &egui::Context,
        output: egui::FullOutput,
    ) -> VkResult<Self> {
        // calculate size for staging buffer.
        // calculate size for staging image.
        // allocate resource ids for textures in tessellated list (imported from texture manager)
        // define accesses for resource ids
        for (egui_id, delta) in output
            .textures_delta
            .set
            .iter()
            .filter(|(_, image)| image.is_whole())
        {
            let image = Image::new(
                dev.clone(),
                ImageDesc {
                    name: Some(format!("egui-texture-{egui_id:?}").into()),
                    format: vk::Format::R8G8B8A8_UNORM,
                    extent: vk::Extent3D {
                        width: delta.image.width() as u32,
                        height: delta.image.height() as u32,
                        depth: 1,
                    },
                    usage: vk::ImageUsageFlags::SAMPLED | vk::ImageUsageFlags::TRANSFER_DST,
                    mem_usage: vk_mem::MemoryUsage::AutoPreferDevice,
                    ..Default::default()
                },
            )?;

            let tid = textures.insert_image(Arc::new(image));
            egui_state.textures.insert(
                *egui_id,
                crate::EguiTextureInfo {
                    id: tid,
                    options: delta.options,
                },
            );
        }

        let (staging_size, image_size) = output.textures_delta.set.iter().fold(
            (0usize, 0usize),
            |(mut buffer, mut image), (_id, delta)| {
                let bytes =
                    delta.image.height() * delta.image.width() * delta.image.bytes_per_pixel();
                if !delta.is_whole() {
                    image = image.max(bytes);
                }
                buffer = buffer + bytes;

                (buffer, image)
            },
        );

        let tessellated = egui.tessellate(output.shapes, output.pixels_per_point);

        let mut staging_buffer = Buffer::new(
            dev.clone(),
            BufferDesc {
                name: Some("egui-prepass-staging-buffer".into()),
                size: staging_size as u64,
                usage: vk::BufferUsageFlags::TRANSFER_SRC,
                queue_families: device::QueueFlags::empty(),
                mem_usage: vk_mem::MemoryUsage::AutoPreferHost,
                alloc_flags: vk_mem::AllocationCreateFlags::MAPPED
                    | vk_mem::AllocationCreateFlags::HOST_ACCESS_SEQUENTIAL_WRITE
                    | vk_mem::AllocationCreateFlags::STRATEGY_FIRST_FIT,
                ..Default::default()
            },
        )?;
        let staging_image = Arc::new(Image::new(
            dev.clone(),
            ImageDesc {
                name: Some("egui-prepass-staging-buffer".into()),
                format: vk::Format::R8G8B8A8_UNORM,
                extent: vk::Extent3D {
                    width: (image_size / 2) as u32,
                    height: (image_size - (image_size / 2)) as u32,
                    depth: 1,
                },
                usage: vk::ImageUsageFlags::TRANSFER_SRC | vk::ImageUsageFlags::TRANSFER_DST,
                queue_families: device::QueueFlags::empty(),
                mem_usage: vk_mem::MemoryUsage::AutoPreferDevice,
                ..Default::default()
            },
        )?);

        let aliased_images = {
            let mut staging_map = staging_buffer.map()?;
            let mut offset = 0;

            let aliased_images = output
                .textures_delta
                .set
                .iter()
                .filter_map(|(id, delta)| {
                    let bytes =
                        delta.image.height() * delta.image.width() * delta.image.bytes_per_pixel();

                    let mem = &mut staging_map[offset..offset + bytes];
                    match &delta.image {
                        egui::ImageData::Color(arc) => {
                            let slice = unsafe {
                                core::slice::from_raw_parts(
                                    arc.pixels.as_ptr().cast::<u8>(),
                                    arc.pixels.len() * size_of::<Color32>(),
                                )
                            };
                            mem[..slice.len()].copy_from_slice(slice);
                        }
                        egui::ImageData::Font(font_image) => {
                            for (i, c) in font_image.srgba_pixels(None).enumerate() {
                                let bytes = c.to_array();
                                mem[i * 4..(i + 1) * 4].copy_from_slice(&bytes);
                            }
                        }
                    }

                    let old_offset = offset;
                    offset += bytes;

                    if !delta.is_whole() {
                        unsafe {
                            let alias = staging_image
                                .clone()
                                .get_alias(ImageDesc {
                                    name: Some(
                                        format!("egui-prepass-staging-aliased-{id:?}").into(),
                                    ),
                                    format: vk::Format::R8G8B8A8_UNORM,
                                    extent: vk::Extent3D {
                                        width: delta.image.width() as u32,
                                        height: delta.image.height() as u32,
                                        depth: 1,
                                    },
                                    usage: vk::ImageUsageFlags::TRANSFER_SRC
                                        | vk::ImageUsageFlags::TRANSFER_DST,
                                    queue_families: device::QueueFlags::empty(),
                                    ..Default::default()
                                })
                                .unwrap();
                            Some((*id, (old_offset, bytes, delta.pos.unwrap(), alias)))
                        }
                    } else {
                        None
                    }
                })
                .collect::<BTreeMap<_, _>>();

            aliased_images
        };

        let out_resources = tessellated
            .iter()
            .filter_map(|prim| {
                if let egui::epaint::Primitive::Mesh(ref mesh) = prim.primitive {
                    Some(mesh.texture_id)
                } else {
                    None
                }
            })
            .dedup()
            .filter_map(|egui_id| {
                egui_state
                    .lookup_texture(egui_id)
                    .and_then(|tid| textures.get_texture(tid).map(|image| (egui_id, tid, image)))
            })
            .map(|(egui_id, tid, image)| {
                let rid = rg.import_image(
                    image,
                    ResourceAccess {
                        stage: vk::PipelineStageFlags2::NONE,
                        mask: vk::AccessFlags2::empty(),
                        layout: Some(vk::ImageLayout::GENERAL),
                    },
                );

                (
                    egui_id,
                    (
                        rid,
                        tid,
                        ResourceAccess {
                            stage: vk::PipelineStageFlags2::TRANSFER,
                            mask: vk::AccessFlags2::TRANSFER_WRITE,
                            layout: Some(vk::ImageLayout::TRANSFER_DST_OPTIMAL),
                        },
                    ),
                )
            })
            .collect::<BTreeMap<_, _>>();

        Ok(Self {
            aliased_images,
            staging_buffer: Arc::new(staging_buffer),
            staging_image,
            out_resources,
            tessellated,
            texture_data: output
                .textures_delta
                .set
                .into_iter()
                .collect::<BTreeMap<_, _>>(),
        })
    }
}

impl Pass for EguiPrePass {
    fn get_read_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
        unimplemented!()
    }

    fn get_write_resource_access(&self, id: RenderGraphResourceId) -> ResourceAccess {
        self.out_resources
            .iter()
            .find(|(_, (id2, _, _))| *id2 == id)
            .map(|(_, (_, _, access))| *access)
            .unwrap()
    }

    fn get_queue_capability_requirements(&self) -> device::QueueFlags {
        device::QueueFlags::TRANSFER
    }

    fn get_read_dependencies<'a>(&'a self) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
        Box::new([].into_iter())
    }

    fn get_write_dependencies<'a>(
        &'a self,
    ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
        Box::new(self.out_resources.values().map(|(id, _, _)| *id))
    }

    fn record(&self, ctx: &RenderContext) -> crate::Result<()> {
        let buffer: Barrier = buffer_barrier(
            self.staging_buffer.handle(),
            0,
            self.staging_buffer.len(),
            ResourceAccess {
                stage: vk::PipelineStageFlags2::NONE,
                mask: vk::AccessFlags2::empty(),
                layout: None,
            },
            ResourceAccess {
                stage: vk::PipelineStageFlags2::TRANSFER,
                mask: vk::AccessFlags2::TRANSFER_READ,
                layout: None,
            },
            None,
        )
        .into();

        unsafe {
            ctx.device
                .dev()
                .cmd_pipeline_barrier2(ctx.cmd.buffer(), &((&buffer).into()));
        }

        for (id, (offset, _, pos, alias)) in &self.aliased_images {
            let image: Barrier = image_barrier(
                alias.handle(),
                alias.format(),
                ResourceAccess {
                    stage: vk::PipelineStageFlags2::NONE,
                    mask: vk::AccessFlags2::empty(),
                    layout: Some(vk::ImageLayout::UNDEFINED),
                },
                ResourceAccess {
                    stage: vk::PipelineStageFlags2::TRANSFER,
                    mask: vk::AccessFlags2::TRANSFER_WRITE,
                    layout: Some(vk::ImageLayout::TRANSFER_DST_OPTIMAL),
                },
                None,
            )
            .into();

            unsafe {
                ctx.device
                    .dev()
                    .cmd_pipeline_barrier2(ctx.cmd.buffer(), &((&image).into()));
            }

            ctx.cmd.copy_buffer_to_image(
                self.staging_buffer.handle(),
                alias.handle(),
                vk::ImageLayout::TRANSFER_DST_OPTIMAL,
                &[vk::BufferImageCopy {
                    buffer_offset: *offset as u64,
                    buffer_row_length: alias.width(),
                    buffer_image_height: alias.height(),
                    image_subresource: vk::ImageSubresourceLayers::default()
                        .aspect_mask(vk::ImageAspectFlags::COLOR)
                        .base_array_layer(0)
                        .mip_level(0)
                        .layer_count(1),
                    image_offset: vk::Offset3D { x: 0, y: 0, z: 0 },
                    image_extent: alias.size(),
                }],
            );

            //self.
            let Some(RenderGraphResource::ImportedImage(texture)) = self
                .out_resources
                .get(id)
                .and_then(|(id, _, _)| ctx.resources.get(id))
            else {
                continue;
            };

            let from_barrier = image_barrier(
                alias.handle(),
                alias.format(),
                ResourceAccess {
                    stage: vk::PipelineStageFlags2::TRANSFER,
                    mask: vk::AccessFlags2::TRANSFER_WRITE,
                    layout: Some(vk::ImageLayout::TRANSFER_DST_OPTIMAL),
                },
                ResourceAccess {
                    stage: vk::PipelineStageFlags2::TRANSFER,
                    mask: vk::AccessFlags2::TRANSFER_READ,
                    layout: Some(vk::ImageLayout::TRANSFER_SRC_OPTIMAL),
                },
                None,
            );
            let to_barrier = image_barrier(
                texture.handle(),
                texture.format(),
                ResourceAccess {
                    stage: vk::PipelineStageFlags2::NONE,
                    mask: vk::AccessFlags2::empty(),
                    layout: Some(vk::ImageLayout::GENERAL),
                },
                ResourceAccess {
                    stage: vk::PipelineStageFlags2::TRANSFER,
                    mask: vk::AccessFlags2::TRANSFER_WRITE,
                    layout: Some(vk::ImageLayout::TRANSFER_DST_OPTIMAL),
                },
                None,
            );

            unsafe {
                ctx.device.dev().cmd_pipeline_barrier2(
                    ctx.cmd.buffer(),
                    &vk::DependencyInfo::default()
                        .image_memory_barriers(&[from_barrier, to_barrier]),
                );
            }
            ctx.cmd.copy_images(
                alias.handle(),
                vk::ImageLayout::TRANSFER_SRC_OPTIMAL,
                texture.handle(),
                vk::ImageLayout::TRANSFER_DST_OPTIMAL,
                &[vk::ImageCopy {
                    src_subresource: vk::ImageSubresourceLayers::default()
                        .aspect_mask(vk::ImageAspectFlags::COLOR)
                        .base_array_layer(0)
                        .mip_level(0)
                        .layer_count(1),
                    src_offset: vk::Offset3D { x: 0, y: 0, z: 0 },
                    dst_subresource: vk::ImageSubresourceLayers::default()
                        .aspect_mask(vk::ImageAspectFlags::COLOR)
                        .base_array_layer(0)
                        .mip_level(0)
                        .layer_count(1),
                    dst_offset: vk::Offset3D {
                        x: pos[0] as i32,
                        y: pos[1] as i32,
                        z: 0,
                    },
                    extent: alias.size(),
                }],
            );
        }

        Ok(())
    }
}

#[derive(Debug)]
pub struct EguiRenderPass {
    descriptor_set: vk::DescriptorSet,
    vertices: Buffer,
    indices: Buffer,
    draw_calls: Buffer,
    texture_ids: Buffer,
    textures_to_free: Vec<crate::texture::TextureId>,
    num_draw_calls: usize,
}

impl Pass for EguiRenderPass {
    fn get_read_resource_access(&self, id: RenderGraphResourceId) -> ResourceAccess {
        todo!()
    }

    fn get_write_resource_access(&self, id: RenderGraphResourceId) -> ResourceAccess {
        todo!()
    }

    fn get_queue_capability_requirements(&self) -> device::QueueFlags {
        todo!()
    }

    fn get_read_dependencies<'a>(&'a self) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
        todo!()
    }

    fn get_write_dependencies<'a>(
        &'a self,
    ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
        todo!()
    }

    fn record(&self, ctx: &RenderContext) -> crate::Result<()> {
        todo!()
    }
}

pub trait Pass: Debug + Send {
    fn get_read_resource_access(&self, id: RenderGraphResourceId) -> ResourceAccess;
    fn get_write_resource_access(&self, id: RenderGraphResourceId) -> ResourceAccess;
    /// mask of the queue capability requirements of this pass.
    fn get_queue_capability_requirements(&self) -> device::QueueFlags;
    /// returns an iterator over all (in) dependencies.
    fn get_read_dependencies<'a>(&'a self) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a>;
    fn get_write_dependencies<'a>(&'a self)
        -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a>;

    fn record(&self, ctx: &RenderContext) -> crate::Result<()>;
}

def_monotonic_id!(pub RenderGraphPassId);

// Non-imported resources remain `RenderGraphResourceDesc`s because they may be
// able to be aliased.
// This should be dual to liveness/register allocation in a compiler.
// Dummy-impl is just allocating every resource_desc itself. 5head-impl is trying
// to find resource_descs which are eq, but whose liveness doesn't overlap.
#[derive(Debug)]
pub struct RenderGraph {
    resource_descs: BTreeMap<RenderGraphResourceId, RenderGraphResourceDesc>,
    resources: BTreeMap<RenderGraphResourceId, RenderGraphResource>,
    accesses: BTreeMap<RenderGraphResourceId, ResourceAccess>,
    passes: Vec<Box<dyn Pass>>,
    /// the rendergraph produces these resources. Any passes on which these
    /// outputs do not depend are pruned.
    outputs: Vec<RenderGraphResourceId>,
}

impl RenderGraph {
    pub fn new() -> Self {
        Self {
            resource_descs: BTreeMap::new(),
            resources: BTreeMap::new(),
            passes: Vec::new(),
            accesses: BTreeMap::new(),
            outputs: Vec::new(),
        }
    }

    pub fn add_resource(&mut self, desc: RenderGraphResourceDesc) -> RenderGraphResourceId {
        let id = RenderGraphResourceId::new();
        self.resource_descs.insert(id, desc);
        self.accesses.insert(id, ResourceAccess::undefined());
        id
    }
    pub fn mark_as_output(&mut self, id: RenderGraphResourceId) {
        // TODO: dedup
        self.outputs.push(id);
    }
    pub fn import_image(
        &mut self,
        image: Arc<Image>,
        access: ResourceAccess,
    ) -> RenderGraphResourceId {
        let id = RenderGraphResourceId::new();
        self.resources
            .insert(id, RenderGraphResource::ImportedImage(image));
        self.accesses.insert(id, access);
        id
    }
    pub fn import_buffer(
        &mut self,
        buffer: Arc<Buffer>,
        access: ResourceAccess,
    ) -> RenderGraphResourceId {
        let id = RenderGraphResourceId::new();
        self.resources
            .insert(id, RenderGraphResource::ImportedBuffer(buffer));
        self.accesses.insert(id, access);
        id
    }
    pub fn import_framebuffer(&mut self, frame: Arc<SwapchainFrame>) -> RenderGraphResourceId {
        let id = RenderGraphResourceId::new();
        self.resources
            .insert(id, RenderGraphResource::Framebuffer(frame));
        id
    }
    pub fn add_pass<P: Pass + 'static>(&mut self, pass: P) {
        self.passes.push(Box::new(pass));
    }

    // https://blog.traverseresearch.nl/render-graph-101-f42646255636
    // https://github.com/EmbarkStudios/kajiya/blob/main/crates/lib/kajiya-rg/src/graph.rs
    // https://themaister.net/blog/2017/08/15/render-graphs-and-vulkan-a-deep-dive/
    pub fn resolve(mut self, device: device::Device) -> crate::Result<()> {
        // create internal resources:
        for (&id, desc) in self.resource_descs.iter() {
            match desc.clone() {
                RenderGraphResourceDesc::Image(image_desc) => {
                    self.resources.insert(
                        id,
                        RenderGraphResource::Image(Image::new(device.clone(), image_desc)?),
                    );
                }
                RenderGraphResourceDesc::Buffer(buffer_desc) => {
                    self.resources.insert(
                        id,
                        RenderGraphResource::Buffer(Buffer::new(device.clone(), buffer_desc)?),
                    );
                }
            }
        }

        eprintln!("{:#?}", &self);
        let mut last_write = BTreeMap::new();

        let mut dag = petgraph::stable_graph::StableDiGraph::new();

        // insert edges between write->read edges of 2 passes
        for (i, pass) in self.passes.iter().enumerate() {
            let node = dag.add_node(i);
            for dep in pass.get_read_dependencies() {
                if let Some(&other) = last_write.get(&dep) {
                    dag.add_edge(other, node, dep);
                }
            }

            for dep in pass.get_write_dependencies() {
                // keep track of which pass last wrote to a resource
                // this is the node to build an edge from next time this resource is read
                last_write.insert(dep, node);
            }
        }

        // pseudo pass for trackingoutputs
        let output = dag.add_node(!0);
        for (id, node) in self
            .outputs
            .iter()
            .filter_map(|&id| last_write.get(&id).cloned().map(|node| (id, node)))
        {
            dag.add_edge(node, output, id);
        }

        // prune dead nodes
        loop {
            let sinks = dag
                .externals(petgraph::Direction::Outgoing)
                .filter(|idx| idx != &output)
                .collect::<Vec<_>>();
            if sinks.is_empty() {
                break;
            }
            for sink in sinks {
                dag.remove_node(sink);
            }
        }

        #[cfg(any(debug_assertions, test))]
        std::fs::write(
            "render_graph.dot",
            &format!(
                "{:?}",
                petgraph::dot::Dot::with_attr_getters(
                    &dag,
                    &[],
                    &|_graph, edgeref| { format!("label = \"{:?}\"", edgeref.weight()) },
                    &|_graph, noderef| { format!("label = \"Pass({:?})\"", noderef.weight()) }
                )
            ),
        )
        .expect("writing render_graph repr");

        struct RenderGraphStage {
            passes: Vec<Box<dyn Pass>>,
            accesses: Vec<(RenderGraphResourceId, ResourceAccess)>,
        }

        let mut topological_map = Vec::new();
        let mut top_dag = dag.clone();
        loop {
            let (sources, indices): (Vec<_>, Vec<_>) = top_dag
                .externals(petgraph::Direction::Incoming)
                .filter(|&id| id != output)
                .filter_map(|id| top_dag.node_weight(id).cloned().map(|idx| (id, idx)))
                .unzip();

            if sources.is_empty() {
                break;
            }

            for &source in &sources {
                top_dag.remove_node(source);
            }

            let mut accesses = BTreeMap::<RenderGraphResourceId, ResourceAccess>::new();

            // TODO: distinguish between make-available and make-visible
            for &idx in &indices {
                let pass = &self.passes[idx];
                for id in pass.get_read_dependencies() {
                    let access = pass.get_read_resource_access(id);
                    accesses
                        .entry(id)
                        .and_modify(|entry| {
                            entry.stage = entry.stage.max(access.stage);
                            entry.mask |= access.mask;
                            // TODO: layout
                        })
                        .or_insert(access);
                }
            }

            topological_map.push((indices, accesses));
        }

        // I don't think this can currently happen with the way passes are added.
        top_dag.remove_node(output);
        if dag.node_count() > 0 {
            panic!("dag is cyclic!");
        }

        eprintln!("topology: {:?}", topological_map);

        let pool =
            commands::SingleUseCommandPool::new(device.clone(), device.graphics_queue().clone())?;

        let resources = &self.resources;
        let cmds = topological_map
            .iter()
            .map(|(set, accesses)| {
                let pool = pool.clone();
                let device = device.clone();
                let passes = set
                    .into_iter()
                    .map(|i| self.passes.remove(*i))
                    .collect::<Vec<_>>();

                let cmd = pool.alloc()?;

                // transitions
                for (&id, &access) in accesses.iter() {
                    Self::transition_resource_to(
                        &mut self.accesses,
                        resources,
                        device.dev(),
                        unsafe { &cmd.buffer() },
                        id,
                        access,
                    );
                }

                let ctx = RenderContext {
                    device,
                    cmd,
                    resources,
                };

                for pass in &passes {
                    pass.record(&ctx)?;
                }

                crate::Result::Ok(ctx.cmd)
            })
            .collect::<crate::Result<Vec<_>>>()?;

        Ok(())
    }

    fn transition_resource_to(
        accesses: &mut BTreeMap<RenderGraphResourceId, ResourceAccess>,
        resources: &BTreeMap<RenderGraphResourceId, RenderGraphResource>,
        dev: &ash::Device,
        cmd: &vk::CommandBuffer,
        id: RenderGraphResourceId,
        access: ResourceAccess,
    ) {
        let old_access = accesses.get(&id);
        let res = resources.get(&id);

        if let (Some(&old_access), Some(res)) = (old_access, res) {
            let barrier: Barrier = match res {
                RenderGraphResource::Framebuffer(arc) => {
                    image_barrier(arc.image, arc.format, old_access, access, None).into()
                }
                RenderGraphResource::ImportedImage(arc) => {
                    image_barrier(arc.handle(), arc.format(), old_access, access, None).into()
                }
                RenderGraphResource::ImportedBuffer(arc) => {
                    buffer_barrier(arc.handle(), 0, arc.len(), old_access, access, None).into()
                }
                RenderGraphResource::Image(image) => {
                    image_barrier(image.handle(), image.format(), old_access, access, None).into()
                }
                RenderGraphResource::Buffer(buffer) => {
                    buffer_barrier(buffer.handle(), 0, buffer.len(), old_access, access, None)
                        .into()
                }
            };

            accesses.insert(id, access);

            unsafe {
                dev.cmd_pipeline_barrier2(*cmd, &((&barrier).into()));
            }
        }
    }
}

enum Barrier {
    Image(vk::ImageMemoryBarrier2<'static>),
    Buffer(vk::BufferMemoryBarrier2<'static>),
}

impl<'a> From<&'a Barrier> for vk::DependencyInfo<'a> {
    fn from(value: &'a Barrier) -> Self {
        let info = vk::DependencyInfo::default();
        let info = match value {
            Barrier::Image(barrier) => info.image_memory_barriers(core::slice::from_ref(barrier)),
            Barrier::Buffer(barrier) => info.buffer_memory_barriers(core::slice::from_ref(barrier)),
        };

        info
    }
}

impl From<vk::ImageMemoryBarrier2<'static>> for Barrier {
    fn from(value: vk::ImageMemoryBarrier2<'static>) -> Self {
        Self::Image(value)
    }
}
impl From<vk::BufferMemoryBarrier2<'static>> for Barrier {
    fn from(value: vk::BufferMemoryBarrier2<'static>) -> Self {
        Self::Buffer(value)
    }
}

pub fn buffer_barrier(
    buffer: vk::Buffer,
    offset: u64,
    size: u64,
    before: ResourceAccess,
    after: ResourceAccess,
    queue_families: Option<(u32, u32)>,
) -> vk::BufferMemoryBarrier2<'static> {
    vk::BufferMemoryBarrier2::default()
        .buffer(buffer)
        .offset(offset)
        .size(size)
        .src_access_mask(before.mask)
        .src_stage_mask(before.stage)
        .dst_access_mask(after.mask)
        .dst_stage_mask(after.stage)
        .src_queue_family_index(
            queue_families
                .map(|(src, _)| src)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
        .dst_queue_family_index(
            queue_families
                .map(|(_, dst)| dst)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
}

pub fn image_barrier(
    image: vk::Image,
    format: vk::Format,
    before_access: ResourceAccess,
    after_access: ResourceAccess,
    queue_families: Option<(u32, u32)>,
) -> vk::ImageMemoryBarrier2<'static> {
    vk::ImageMemoryBarrier2::default()
        .src_access_mask(before_access.mask)
        .src_stage_mask(before_access.stage)
        .dst_access_mask(after_access.mask)
        .dst_stage_mask(after_access.stage)
        .image(image)
        .old_layout(before_access.layout.unwrap_or_default())
        .new_layout(after_access.layout.unwrap_or_default())
        .subresource_range(vk::ImageSubresourceRange {
            aspect_mask: util::image_aspect_from_format(format),
            ..images::SUBRESOURCERANGE_ALL
        })
        .src_queue_family_index(
            queue_families
                .map(|(src, _)| src)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
        .dst_queue_family_index(
            queue_families
                .map(|(_, dst)| dst)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
}

#[cfg(test)]
mod tests {
    use super::*;

    macro_rules! def_dummy_pass {
        ($name:ident: {$queue:path, $layout_in:path, $layout_out:path}) => {
            #[derive(Debug, Clone)]
            struct $name(Vec<RenderGraphResourceId>, Vec<RenderGraphResourceId>);
            impl Pass for $name {
                fn get_read_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
                    ResourceAccess::default()
                }
                fn get_write_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
                    ResourceAccess::default()
                }

                fn get_queue_capability_requirements(&self) -> device::QueueFlags {
                    $queue
                }

                fn get_read_dependencies<'a>(
                    &'a self,
                ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
                    Box::new(self.0.iter().cloned())
                }

                fn get_write_dependencies<'a>(
                    &'a self,
                ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
                    Box::new(self.1.iter().cloned())
                }

                fn record(&self, _ctx: &RenderContext) -> crate::Result<()> {
                    Ok(())
                }
            }
        };
    }

    def_dummy_pass!(DepthPass: {
        device::QueueFlags::GRAPHICS,
        vk::ImageLayout::DEPTH_ATTACHMENT_OPTIMAL,
        vk::ImageLayout::DEPTH_ATTACHMENT_OPTIMAL});
    def_dummy_pass!(RenderPass: {
        device::QueueFlags::GRAPHICS,
        vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL,
        vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL});
    def_dummy_pass!(AsyncPass: {
        device::QueueFlags::ASYNC_COMPUTE,
        vk::ImageLayout::UNDEFINED,
        vk::ImageLayout::GENERAL});
    def_dummy_pass!(PostProcessPass: {
        device::QueueFlags::ASYNC_COMPUTE,
        vk::ImageLayout::GENERAL,
        vk::ImageLayout::GENERAL});
    def_dummy_pass!(PresentPass: {
        device::QueueFlags::PRESENT,
        vk::ImageLayout::PRESENT_SRC_KHR,
        vk::ImageLayout::UNDEFINED});
    def_dummy_pass!(DepthVisualisationPass: {
        device::QueueFlags::ASYNC_COMPUTE,
        vk::ImageLayout::GENERAL,
        vk::ImageLayout::UNDEFINED});

    #[test]
    fn resolve_graph() {
        let mut graph = RenderGraph::new();
        let gbuffer = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
            ..Default::default()
        }));
        let depth_image = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
            ..Default::default()
        }));
        let depth_visualisation = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
            ..Default::default()
        }));
        let compute_buffer = graph.add_resource(RenderGraphResourceDesc::Buffer(BufferDesc {
            ..Default::default()
        }));

        graph.add_pass(DepthPass(vec![depth_image], vec![depth_image]));
        graph.add_pass(DepthVisualisationPass(
            vec![depth_image, depth_visualisation],
            vec![depth_visualisation],
        ));
        graph.add_pass(AsyncPass(vec![compute_buffer], vec![compute_buffer]));
        graph.add_pass(RenderPass(
            vec![depth_image, compute_buffer, gbuffer],
            vec![gbuffer],
        ));
        graph.add_pass(PostProcessPass(vec![gbuffer], vec![gbuffer]));
        graph.mark_as_output(gbuffer);
        graph.mark_as_output(depth_image);

        // graph.resolve();
    }
}
